import { Component, Input, OnChanges, SimpleChanges } from '@angular/core';
import { Font } from '@nativescript/core';

import { Align } from '@nativescript-community/ui-canvas';

import { BarChart } from '@nativescript-community/ui-chart/charts/BarChart';
import { BarData } from '@nativescript-community/ui-chart/data/BarData';
import { BarDataSet } from '@nativescript-community/ui-chart/data/BarDataSet';

import { XAxis, XAxisPosition } from '@nativescript-community/ui-chart/components/XAxis';
import { XAxisFormatter } from './formatter/x-axis-formatter';

import { FontWeight } from '@nativescript/core/ui/styling/font';
import { Chart } from '../../models/chart-types';
import { ColorLib } from '../../utils/color-lib';
import { IChart } from './model/IChart';

/**
 * Chart Component
 */
@Component({
  selector: 'ac-charts',
  templateUrl: './charts.component.html',
  styleUrls: ['./charts.component.scss']
})
export class ChartsComponent implements OnChanges {
  /**
   * Used to define type of the chart
   */
  @Input() type: Chart = 'donut';

  /**
   * Data source for the chart
   */
  @Input() data: IChart[];

  /**
   * Width of the chart
   */
  @Input() width: number = 160;

  /**
   * Height of the chart
   */
  @Input() height: number = 160;

  /**
   * Label for the chart
   */
  @Input() label: string = '';

  /**
   * In the data, what property name to be used as chart X Values
   */
  @Input() propertyXName: string;

  /**
   * In the data, what property name to be used as chart Y Values
   */
  @Input() propertyYName: string;

  /**
   * Bar chart reference
   */
  private barRef: BarChart;

  /**
   * Typeface for chart legends
   */
  private readonly baseTypeFace = new Font('Source Sans Pro', 12, 'normal', FontWeight.NORMAL);

  private readonly animationTime = 750;

  /**
   * Gets the x axis.
   *
   * @returns The x axis.
   */
  private get xAxis(): XAxis {
    const chartReference = this.barRef;
    return chartReference.getXAxis();
  }

  /**
   * Angular lifecycle method
   *
   * @param changes latest input changes
   * @param changes.data we are destructuring the data
   */
  ngOnChanges({ data }: SimpleChanges): void {
    if (this.validateChange(data)) {
      if (this.barRef) this.createBarChart();
    }
  }

  /**
   * Creates a bar chart reference
   *
   * @param event object
   * @param event.object bar chart object
   */
  public onBarChartLoaded({ object: barChart }: { object: BarChart }): void {
    this.barRef = barChart;
    this.createBarChart();
  }

  /**
   * Creates a bar chart
   */
  public createBarChart(): void {
    this.barRef.setClipValuesToContent(false);
    this.styleChartAxis();
    this.barRef.setScaleEnabled(false);

    const barDataSet = new BarDataSet(this.data, this.label, this.propertyXName, this.propertyYName);

    const barColors = this.getChartDataWithColors();

    barDataSet.setColors(barColors);
    barDataSet.setValueTypeface(this.baseTypeFace.withFontSize(12).withFontWeight(FontWeight.SEMI_BOLD));
    barDataSet.setDrawValues(true);

    const barDataSets: BarDataSet[] = [barDataSet];

    const barData = new BarData(barDataSets);
    barData.setDrawValues(true);

    /**
     * This is the bar width in "values" according to the documentation. I determined this number by considering a bar and
     * the space next to it as "one value". In the design, the bars are 66px and the space next to them is 24, which is
     * 90. I then took 66/90 and got 0.73 "values".
     *
     * @see https://github.com/nativescript-community/ui-chart/blob/master/src/charting/data/BarData.ts#L17
     */
    barData.setBarWidth(0.73);

    this.barRef.animateXY(this.animationTime, this.animationTime);
    this.barRef.setData(barData);
  }
  /**
   * Combines the chart data and the color template
   *
   * @returns Array of color values
   */
  private getChartDataWithColors(): string[] {
    return this.data.reduce((colors: string[], item): string[] => {
      return [...colors, ColorLib.getColorForIndex(item.index)];
    }, []);
  }

  /**
   * will get list of labels for the corresponding value
   *
   * @param xName property to read
   * @returns object consists of key label value and the actual labels
   */
  private getLabels(xName: string): { [key: number]: string } {
    return this.data.reduce((labels: object, data: IChart) => {
      return { ...labels, [data[xName]]: data.label };
    }, {});
  }

  /**
   * utility function to validate change on a variable
   *
   * @param updatedVal update data
   * @param updatedVal.firstChange is the change happens first
   * @param updatedVal.previousValue it holds the previous value
   * @param updatedVal.currentValue it has updated value
   * @returns is it pass the condition
   */
  private validateChange({ firstChange, previousValue, currentValue }): boolean {
    return !firstChange && previousValue !== currentValue;
  }

  /**
   * Styles the axis for charts other than line charts
   */
  private styleChartAxis(): void {
    const axisLineColor = '#CCC';
    const axisTextColor = '#999';
    const drawGridLines = false;
    const textSize = 12;

    this.xAxis.setEnabled(false);
    this.xAxis.setDrawLabels(true);
    this.xAxis.setPosition(XAxisPosition.BOTTOM);
    this.xAxis.setLabelCount(this.data.length);
    this.xAxis.setLabelTextAlign(Align.CENTER);
    this.xAxis.setValueFormatter(new XAxisFormatter(this.getLabels(this.propertyXName)));

    const yAxis = this.barRef.axisLeft;
    yAxis.setDrawGridLines(drawGridLines);
    yAxis.setDrawAxisLine(0);
    yAxis.setTextColor(axisTextColor);
    yAxis.setTextSize(textSize);
    yAxis.setDrawZeroLine(true);
    yAxis.setZeroLineColor(axisLineColor);

    const yValues = this.data.map((dataPoint: IChart) => Number(dataPoint.value));

    const interval = 5;
    yAxis.setForcedInterval(interval);
    yAxis.setAxisMinimum(Math.min(...yValues) - interval);
    yAxis.setAxisMaximum(Math.max(...yValues) + interval);
  }
}

